# xdr_stories.py
#
#
# Version: 1.0.0
# Author: Peter Lee, April 2024
#
#
# This script takes as input an API key, account ID, and optional timeframe, and returns XDR stories
# updated during that timeframe. 
#
# Usage: xdr_stories.py [options]
#
# Options:
#   -h, --help    show this help message and exit
#   -K API_KEY    API key
#   -I ID         Account ID
#   -P            Prettify output
#   -p            Print stories
#   -t TIMEFRAME  timeframe. Default is last 60 minutes (last.PT60M)
#   -v            Print debug info
#   -V            Print detailed debug info
#
# Examples:
#
# To run the script with key=YOURAPIKEY for account ID 1714, fetching stories for the last hour:
#   python3 xdr_stories.py -K YOURAPIKEY -I 1714
#
# Running the script with debug enabled so you can see the fetch logic in action:
#   python3 xdr_stories.py -K YOURAPIKEY -I 1714 -v
#
# To show stories, use the -p parameter:
#   python3 xdr_stories.py -K YOURAPIKEY -I 1714 -p
#
# For more human readable stories, use -pP
#   python3 xdr_stories.py -K YOURAPIKEY -I 1714 -pP
#
# This script is supplied as a demonstration of how to access the Cato API with Python. It
# is not an official Cato release and is provided with no guarantees of support. Error handling
# is restricted to the bare minimum required for the script to work with the API, and may not be
# sufficient for production environments.
#
# All questions or feedback should be sent to api@catonetworks.com

import base64
import datetime
import gzip
import hmac
import hashlib
import json
import os
import socket
import ssl
import sys
import time
import urllib.parse
import urllib.request
from optparse import OptionParser


########################################################################################
########################################################################################
########################################################################################
# Helper functions and globals

# log debug output
def log(text):
    if options.verbose or options.veryverbose:
        print(f"LOG {datetime.datetime.now(datetime.UTC)}> {text}")

# log detailed debug output
def logd(text):
    if options.veryverbose:
        log(text)


# send GQL query string to API, return JSON
# if we hit a network error, retry ten times with a 2 second sleep
def send(query):
    global api_call_count
    global total_bytes_compressed
    global total_bytes_uncompressed
    retry_count = 0
    data = {'query':query}
    headers = { 'x-api-key': options.api_key,'Content-Type':'application/json', 'Accept-Encoding':'gzip, deflate, br'}
    no_verify = ssl._create_unverified_context()
    while True:
        if retry_count > 10:
            print("FATAL ERROR retry count exceeded")
            sys.exit(1)
        try:
            request = urllib.request.Request(url='https://api.catonetworks.com/api/v1/graphql2',
                data=json.dumps(data).encode("ascii"),headers=headers)
            response = urllib.request.urlopen(request, context=no_verify, timeout=30)
            api_call_count += 1
        except Exception as e:
            log(f"ERROR {retry_count}: {e}, sleeping 2 seconds then retrying")
            time.sleep(2)
            retry_count += 1
            continue
        zipped_data = response.read()
        total_bytes_compressed += len(zipped_data)
        result_data = gzip.decompress(zipped_data)
        total_bytes_uncompressed += len(result_data)
        if result_data[:48] == b'{"errors":[{"message":"rate limit for operation:':
            log("RATE LIMIT sleeping 5 seconds then retrying")
            time.sleep(5)
            continue
        break
    result = json.loads(result_data.decode('utf-8','replace'))
    if "errors" in result:
        log(f"API error: {result_data}")
        return False,result
    return True,result



########################################################################################
########################################################################################
########################################################################################
# start of the main program

api_call_count = 0
total_bytes_compressed = 0
total_bytes_uncompressed = 0
start = datetime.datetime.now()

# Process options
parser = OptionParser()
parser.add_option("-K", dest="api_key", help="API key")
parser.add_option("-I", dest="ID", help="Account ID")
parser.add_option("-P", dest="prettify", action="store_true", help="Prettify output")
parser.add_option("-p", dest="print_stories", action="store_true", help="Print stories")
parser.add_option("-S", dest="stacked_json", action="store_true", help="Print stacked JSON (one line per story)")
parser.add_option("-t", dest="timeframe", default="last.PT60M", help="timeframe. Default is last 60 minutes (last.PT60M)")
parser.add_option("-v", dest="verbose", action="store_true", help="Print debug info")
parser.add_option("-V", dest="veryverbose", action="store_true", help="Print detailed debug info")
(options, args) = parser.parse_args()
if options.api_key is None or options.ID is None:
    parser.print_help()
    sys.exit(1)


# API call loop
iteration = 1
total_count = 0
stories = []
while True:
    query = '''
{
    xdr(accountID:'''+options.ID+''') {
        stories(
            input: {paging: {from:'''+str(len(stories))+''', limit:100}
            filter: [{timeFrame: {time:"'''+options.timeframe+'''"}}]}
        ) {
            paging {
                from
                limit
                total
            }
            items {
                id
                accountId
                analystName
                analystEmail
                accountName
                updatedAt
                createdAt
                playbook
                summary
                incident {
                    id
                    firstSignal
                    lastSignal
                    engineType
                    vendor
                    producerName
                    connectionType
                    indication
                    queryName
                    criticality
                    source
                    ticket
                    status
                    research
                    siteName
                    storyDuration
                    description
                    sourceIp
                    analystFeedback {
                        verdict
                        severity
                        threatType {
                            name
                            recommendedAction
                            details
                        }
                        threatClassification
                        additionalInfo
                    }
                    site {
                        id
                        name
                    }
                    user {
                        id
                        name
                    }
                    predictedVerdict
                    predictedThreatType
                    ... on ThreatPrevention {
                        similarStoriesData {
                            storyId
                            threatTypeName
                            verdict
                            threatClassification
                            similarityPercentage
                            indication
                        }
                        srcSiteId
                        flowsCardinality
                        riskLevel
                        os
                        deviceName
                        macAddress
                        sourceIp
                        logonName
                        direction
                        events {
                            signatureId
                            eventType
                            threatType
                            threatName
                            severity
                            action
                        }
                        mitres {id name}
                        timeSeries {
                            data
                            label
                            sum
                            units
                            info
                            key {
                                measureFieldName 
                                dimensions {fieldName value}
                            }
                        }
                        targets {
                            type
                            name
                            analysisScore
                            infectionSource
                            threatReference
                            catoPopularity
                            threatFeeds
                            creationTime
                            categories
                            countryOfRegistration
                            searchHits
                            engines
                            eventData {
                                signatureId
                                eventType
                                threatType
                                threatName
                                severity
                                action
                            }
                        }
                        threatPreventionsEvents {
                            appName
                            clientClass
                            sourceIp
                            sourcePort
                            destinationCountry
                            destinationIp
                            destinationPort
                            direction
                            createdAt
                            method
                            url
                            target
                            domain
                            sourceGeolocation
                            destinationGeolocation
                            tunnelGeolocation
                            dnsResponseIP
                            smbFileName
                            user
                            userAgent
                            fileHash
                            ja3
                            referrer
                            httpResponseCode
                        }
                    }
                    ... on CatoEndpoint {
                        similarStoriesData {
                            storyId
                            threatTypeName
                            verdict
                            threatClassification
                            similarityPercentage
                            indication
                        }
                        device {
                            id
                            deviceName
                            osDetails {
                                osType
                                osBuild
                                osVersion
                            }
                            loggedOnUsers {
                                id
                                name
                                ... on MicrosoftEndpointUser {
                                    userSid
                                    accountName
                                    domainName
                                    principalName
                                }
                            }
                            macAddress
                        }
                    }
                    ... on AnomalyStats {
                        similarStoriesData {
                            storyId
                            threatTypeName
                            verdict
                            threatClassification
                            similarityPercentage
                            indication
                        }
                        srcSiteId
                        os
                        deviceName
                        macAddress
                        logonName
                        clientClass
                        drillDownFilter {name value}
                        breakdownField
                        subjectType
                        extra {name type value}
                        gaussian {std ss z_score avg n}
                        metric {name value}
                        metricDetails {name units}
                        mitres {id name}
                        rules
                        timeSeries {
                            data
                            label
                            sum
                            units
                            info
                            key {
                                measureFieldName 
                                dimensions {fieldName value}
                            }
                        }
                        targets {
                            type
                            name
                            analysisScore
                            infectionSource
                            threatReference
                            catoPopularity
                            threatFeeds
                            creationTime
                            categories
                            countryOfRegistration
                            searchHits
                            engines
                            eventData {
                                signatureId
                                eventType
                                threatType
                                threatName
                                severity
                                action
                            }
                        }                        
                    }
                    ... on AnomalyEvents {
                        similarStoriesData {
                            storyId
                            threatTypeName
                            verdict
                            threatClassification
                            similarityPercentage
                            indication
                        }
                        srcSiteId
                        os
                        deviceName
                        macAddress
                        logonName
                        clientClass
                        drillDownFilter {name value}
                        breakdownField
                        subjectType
                        extra {name type value}
                        gaussian {std ss z_score avg n}
                        metric {name value}
                        metricDetails {name units}
                        mitres {id name}
                        rules
                        timeSeries {
                            data
                            label
                            sum
                            units
                            info
                            key {
                                measureFieldName 
                                dimensions {fieldName value}
                            }
                        }
                        targets {
                            type
                            name
                            analysisScore
                            infectionSource
                            threatReference
                            catoPopularity
                            threatFeeds
                            creationTime
                            categories
                            countryOfRegistration
                            searchHits
                            engines
                            eventData {
                                signatureId
                                eventType
                                threatType
                                threatName
                                severity
                                action
                            }
                        }                        
                    }
                    ... on Threat {
                        similarStoriesData {
                            storyId
                            threatTypeName
                            verdict
                            threatClassification
                            similarityPercentage
                            indication
                        }
                        srcSiteId
                        flowsCardinality
                        riskLevel
                        os
                        deviceName
                        macAddress
                        sourceIp
                        logonName
                        direction
                        clientClass
                        events {
                            signatureId
                            eventType
                            threatType
                            threatName
                            severity
                            action
                        }                        
                        mitres {id name}
                        timeSeries {
                            data
                            label
                            sum
                            units
                            info
                            key {
                                measureFieldName 
                                dimensions {fieldName value}
                            }
                        }
                        targets {
                            type
                            name
                            analysisScore
                            infectionSource
                            threatReference
                            catoPopularity
                            threatFeeds
                            creationTime
                            categories
                            countryOfRegistration
                            searchHits
                            engines
                            eventData {
                                signatureId
                                eventType
                                threatType
                                threatName
                                severity
                                action
                            }
                        }                        
                        flows {
                            appName
                            clientClass
                            sourceIp
                            sourcePort
                            destinationCountry
                            destinationIp
                            destinationPort
                            direction
                            createdAt
                            referer
                            userAgent
                            method
                            url
                            target
                            domain
                            sourceGeolocation
                            destinationGeolocation
                            tunnelGeolocation
                            httpResponseCode
                            dnsResponseIP
                            smbFileName
                            user
                            fileHash
                            ja3
                        }
                    }  
                    ... on NetworkXDRIncident {
                        similarStoriesData {
                            storyId
                            threatTypeName
                            verdict
                            threatClassification
                            similarityPercentage
                            indication
                        }
                        networkIncidentTimeline {
                            created
                            validated
                            description
                            eventType
                            incidentId
                            networkEventSource
                            eventIds
                            acknowledged
                            linkId
                            linkName
                            linkConfigPrecedence
                            linkStatus
                            linkConfigBandwidth
                            deviceConfigHaRole
                            deviceHaRoleState
                            pop
                            isp
                            bgpConnection {
                                connectionName
                                peerIp
                                peerAsn
                                catoIp
                                catoAsn
                            }
                            linkQualityIssue {
                                issueType
                                direction
                                current
                                threshold
                            }
                            hostIp
                            ruleName
                            tunnelResetCount
                        }
                        storyType
                        occurrences
                        siteConnectionType
                        siteConfigLocation
                        acknowledged
                        linkId
                        linkName
                        linkConfigPrecedence
                        deviceConfigHaRole
                        licenseRegion
                        licenseBandwidth
                        pop
                        isp
                        bgpConnection {
                            connectionName
                            peerIp
                            peerAsn
                            catoIp
                            catoAsn
                        }
                        hostIp
                        ruleName
                    }  
                    ... on MicrosoftEndpoint {
                        similarStoriesData {
                            storyId
                            threatTypeName
                            verdict
                            threatClassification
                            similarityPercentage
                            indication
                        }
                        device {
                            id
                            deviceName
                            osDetails {
                                osType
                                osBuild
                                osVersion
                            }
                            loggedOnUsers {
                                id
                                name
                                ... on MicrosoftEndpointUser {
                                    userSid
                                    accountName
                                    domainName
                                    principalName
                                }
                            }
                            firstSeenDateTime
                            avStatus
                            healthStatus
                            rbacGroup {id name}
                            ipInterfaces
                            azureAdDeviceId
                            onboardingStatus
                        }
                        alerts {
                            id
                            title
                            description
                            threatName
                            mitreTechnique {id name}
                            mitreSubTechnique {id name}
                            createdDateTime
                            resources {
                                id
                                createdDateTime
                                remediationStatus
                                remediationStatusDetails
                                tags
                                roles
                                verdict
                                ... on MicrosoftProcessResource {
                                    processId
                                    processCommandLine
                                    imageFile {
                                        name
                                        path
                                        size
                                        sha1
                                        sha256
                                        md5
                                        issuer
                                        signer
                                        publisher
                                    }
                                    userAccount {
                                        id
                                        name
                                        ... on MicrosoftEndpointUser {
                                            userSid
                                            accountName
                                            domainName
                                            principalName
                                        }
                                    }
                                }
                                ... on MicrosoftFileResource {
                                    fileDetails {
                                        name
                                        path
                                        size
                                        sha1
                                        sha256
                                        md5
                                        issuer
                                        signer
                                        publisher
                                    }
                                    detectionStatus
                                }
                                ... on MicrosoftRegistryResource {
                                    hive
                                    key
                                    value
                                    valueName
                                    valueType
                                }                                    
                            }
                            activities {
                                id
                                resourceId
                                parentResourceId
                                action
                                firstActivityDateTime
                                lastActivityDateTime
                            }
                            criticality
                            comments
                            recommendedActions
                            category
                            ownerName
                            threatFamilyName
                            threatType
                            resolvedDateTime
                            firstActivityDateTime
                            lastActivityDateTime
                            lastUpdateDateTime
                            localIp
                            destinationIp
                            destinationUrl
                            status
                            providerAlertId
                            alertWebUrl
                            determination
                            detectionSource
                            classification
                        }
                    }                                          
                }
                timeline {
                    createdAt
                    description
                    context
                    type
                    descriptions
                    category
                    additionalInfo
                    analystInfo {name email}
                }
            }
        }
    }
}
'''

    logd(query)
    success,resp = send(query)
    if not success:
        print(resp)
        sys.exit(1)
    logd(resp)

    # append to list of stories
    stories += resp["data"]["xdr"]["stories"]["items"]

    # record total
    total = int(resp["data"]["xdr"]["stories"]["paging"]["total"])
    fetched_count = len(resp["data"]["xdr"]["stories"]["items"])
    total_count = len(stories)
    line = f"iteration:{iteration} fetched:{fetched_count} total_fetched:{total_count} target:{total}"
    if fetched_count > 0:
        line += " "+resp["data"]["xdr"]["stories"]["items"][0]["updatedAt"]
        line += " "+resp["data"]["xdr"]["stories"]["items"][-1]["updatedAt"]
    log(line)

    # increment counter and check if we hit any limits for stopping
    iteration += 1
    if len(stories) >= total:
        log(f'Retrieved {len(stories)} from a total of {total}, no more to retrieve')
        break


# print output
if options.print_stories and len(stories) > 0:
    if options.prettify:
        print(json.dumps(stories,indent=2, ensure_ascii=False))
    else:
        if options.stacked_json:
            for story in stories:
                try:
                    print(json.dumps(story, ensure_ascii=False))
                except:
                    print(json.dumps(story))            
        else:
            try:
                print(json.dumps(stories, ensure_ascii=False))
            except:
                print(json.dumps(stories))


end = datetime.datetime.now()
log(f"OK {len(stories)} stories from {api_call_count} API calls with {total_bytes_uncompressed} bytes uncompressed, {total_bytes_compressed} bytes compressed in {end-start}")
